/* file: gqrs.c			G. Moody		16 November 2006
				Last revised:		16 January 2008   
-------------------------------------------------------------------------------
gqrs: A QRS detector
Copyright (C) 2006-2008 George B. Moody

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA.

You may contact the author by e-mail (george@mit.edu) or postal mail
(MIT Room E25-505A, Cambridge, MA 02139 USA).  For updates to this software,
please visit PhysioNet (http://www.physionet.org/).
_______________________________________________________________________________

A future version of this detector will analyze multiple ECG signals from a
single subject.  Much, but not all, of the code necessary to do so is already
included here, but the main() function forces nsig, the number of signals to be
analyzed, to be 1 since the multi-signal code is incomplete.  If the symbol
MULTILEAD is defined at compilation time, this constraint is disabled.
*/

#include <stdio.h>
#include <stddef.h>
#include <wfdb/wfdb.h>
#include <wfdb/ecgcodes.h>

#define BUFLN 2048	/* must be a power of 2, see qf() */
#define NPEAKS 64	/* number of peaks buffered (per signal) */

/* The `getconf' macro is used by gqrs_init() (below) to check a line of input
   (already in `buf', defined in gqrs_init) for the string named by getconf's
   first argument.  If the string is found, the value following the string
   (and an optional `:' or `=') is converted using sscanf and the format
   specifier supplied as getconf's second argument, and stored in the variable
   named by the first argument. */
#define getconf(a, fmt)	if (p=strstr(buf,#a)){sscanf(p,#a "%*[=: \t]" fmt,&a);}

/* The 'peak' structure contains information about a local maximum in the
   filtered signal (qfv).  Peaks are stored in circular buffers of peak
   structures, one buffer per input signal.  The time of a flat-topped peak
   is the time of the first sample that has the maximum value.  A peak is
   secondary if there is a larger peak within its neighborhood (time +- rrmin),
   of if it has been identified as a T-wave associated with a previous primary
   peak.  A peak is primary if it is largest in its neighborhood, or if the
   only larger peaks are secondary. */
struct peak {
    struct peak *prev, *next; /* pointers to neighbors (in time) */
    WFDB_Time time;	/* time of local maximum of qfv */
    int amp;	 /* value of qfv at time of peak */
    short type;  /* 1: primary, 2: secondary, 0: not determined */
};

/* Prototypes of functions defined below.  The definitions of these functions
   follow that of main(), in the order shown below. */
WFDB_Sample sm(WFDB_Signal c, WFDB_Time t);
void qf(void);
void addpeak(WFDB_Signal c, WFDB_Time t, int peak_amplitude);
int peaktype(struct peak *p);
void gqrs_init(WFDB_Time from, WFDB_Time to);
struct peak *find_missing(struct peak *previous_peak, struct peak *next_peak);
void gqrs(WFDB_Time from, WFDB_Time to);
void help(void);
char *prog_name(char *p);
void *gcalloc(size_t nmemb, size_t size);
void cleanup(int status);

char *pname;			/* name of this program, used in messages */
char *record = NULL;		/* name of input record */
char *orec = NULL;		/* name of output record, if any */
char *osfname = NULL;		/* name of output signal file, if any */
FILE *config = NULL;		/* configuration file, if any */
int minutes = 0;		/* minutes elapsed in the current hour */
int nsig = 0;			/* number of signals to be analyzed */
int *pthr;			/* peak-detection threshold */
int *qthr;			/* QRS-detection threshold */
int *pthmin, *qthmin;		/* minimum values for pthr and qthr */
int rrdev;			/* mean absolute deviation of RR from rrmean */
int rrinc;			/* maximum incremental change in rrmean */
int rrmean;			/* mean RR interval, in sample intervals */
int rrmax;			/* maximum likely RR interval */
int rrmin;			/* minimum RR interval, in sample intervals */
int rtmax;			/* maximum RT interval, in sample intervals */
int rtmean;			/* mean RT interval, in sample intervals */
int *tamean;			/* mean T-wave amplitude in qfv */
double thresh = 1.0;		/* normalized detection threshold */
long v1norm;			/* normalization for v1 in qf() */
WFDB_Annotation annot;		/* most recently written annotation */
WFDB_Sample *v;			/* latest sample of each input signal */
WFDB_Sample *vo;		/* latest sample of each output signal */
WFDB_Siginfo *si;		/* characteristics of each input signal */
WFDB_Siginfo *so;		/* characteristics of each output signal */
WFDB_Signal *sig = NULL;	/* signal numbers of signals to be analyzed */
WFDB_Time next_minute;		/* sample number of start of next minute */
WFDB_Time spm;			/* sample intervals per minute */
WFDB_Time sps;			/* sample intervals per second */
WFDB_Time t;			/* time of the current sample being analyzed */

/* Current operating mode of the detector */
enum { LEARNING, RUNNING, CLEANUP } state = LEARNING;

main(int argc, char **argv)
{
    char *p;
    static char auxbuf[1+255+1];
    int gvmode = 0, i, isiglist = 0, j, nisig, s;
    WFDB_Anninfo a;
    WFDB_Time from = (WFDB_Time)0, to = (WFDB_Time)0;

    pname = prog_name(argv[0]);

    for (i = 1; i < argc; i++) {
	if (*argv[i] == '-') switch (*(argv[i]+1)) {
	  case 'c':	/* configuration file */
	    if (++i >= argc) {
		(void)fprintf(stderr,
		     "%s: name of configuration file must follow -c\n", pname);
		cleanup(1);
	    }
	    if ((config = fopen(argv[i], "rt")) == NULL) {
		(void)fprintf(stderr,
		     "%s: can't read configuration file %s\n", pname, argv[i]);
		cleanup(1);
	    }
	    break;
	  case 'f':	/* starting time */
	    if (++i >= argc) {
		(void)fprintf(stderr, "%s: time must follow -f\n", pname);
		cleanup(1);
	    }
	    from = i;
	    break;
	  case 'h':	/* help requested */
	    help();
	    cleanup(0);
	    break;
	  case 'H':	/* operate in WFDB_HIGHRES mode */
	    gvmode = WFDB_HIGHRES;
	    break;
	  case 'm':	/* threshold */
	    if (++i >= argc) {
		(void)fprintf(stderr, "%s: threshold must follow -m\n", pname);
		cleanup(1);
	    }
	    thresh = atof(argv[i]);
	    break;
	  case 'o':	/* write output signals to specified record */
	    if (++i >= argc) {
		(void)fprintf(stderr,"%s: output record name must follow -r\n",
			      pname);
		cleanup(1);
	    }
	    orec = argv[i];
	    break;
	  case 'r':	/* record name */
	    if (++i >= argc) {
		(void)fprintf(stderr, "%s: input record name must follow -r\n",
			      pname);
		cleanup(1);
	    }
	    record = argv[i];
	    break;
	  case 's':	/* signal list follows */
	    isiglist = i+1; /* index of first argument containing a signal # */
	    while (i+1 < argc && *argv[i+1] != '-') {
		i++;
		nsig++;	/* number of elements in signal list */
	    }
	    if (nsig == 0) {
		(void)fprintf(stderr, "%s: signal list must follow -s\n",
			pname);
		cleanup(1);
	    }
	    break;
	  case 't':	/* end time */
	    if (++i >= argc) {
		(void)fprintf(stderr, "%s: time must follow -t\n",pname);
		cleanup(1);
	    }
	    to = i;
	    break;
	  default:
	    (void)fprintf(stderr, "%s: unrecognized option %s\n", pname,
			  argv[i]);
	    cleanup(1);
	}
	else {
	    (void)fprintf(stderr, "%s: unrecognized argument %s\n", pname,
			  argv[i]);
	    cleanup(1);
	}
    }

#ifndef MULTILEAD
    if (nsig > 1) {
	fprintf(stderr,
		"warning: this version of %g analyzes only one signal!\n",
		pname);
	nsig = 1;
    }
#endif

    if (record == NULL) {
	help();
	cleanup(1);
    }

    if (gvmode == 0 && (p = getenv("WFDBGVMODE")))
	gvmode = atoi(p);
    setgvmode(gvmode|WFDB_GVPAD);

    if ((nisig = isigopen(record, NULL, 0)) < 1) cleanup(2);
    si = (WFDB_Siginfo *)gcalloc((size_t)nisig, sizeof(WFDB_Siginfo));
    v = (WFDB_Sample *)gcalloc((size_t)nisig, sizeof(WFDB_Sample));
    a.name = "qrs"; a.stat = WFDB_WRITE;
    if ((nisig = wfdbinit(record, &a, 1, si, nisig)) < 1) cleanup(2);
    if (sampfreq((char *)NULL) < 50.) {
	(void)fprintf(stderr, "%s: sampling frequency (%g Hz) is too low%s",
		      pname, sampfreq((char *)NULL),
		      gvmode & WFDB_HIGHRES ? "\n" : ", try -H option\n");
	cleanup(3);
    }
    if (from > 0L && (from = strtim(argv[from])) < 0L)
	    from = -from;
    if (to > 0L) {
	if ((to = strtim(argv[to])) < 0L)
	    to = -to;
	to += sps;    /* make sure that the last beat before 'to' is marked */
    }
    t = from;
    sps = strtim("1");
    spm = strtim("1:0");
    next_minute = t + spm;

    if (nsig > 0) {	/* analyze specified signals only */
	sig = (WFDB_Signal *)gcalloc((size_t)nsig, sizeof(WFDB_Signal));
	for (i = j = 0; i < nsig; i++) {
	    p = argv[isiglist+i];
	    if ('0' <= *p && *p <= '9') {
		if ((s = atoi(p)) >= nisig)
		    (void)fprintf(stderr,
				  "%s: (warning) no signal %d in record %s\n",
				  pname, s, record);
	    }
	    else {
		for (s = 0; s < nisig; s++)
		    if (strncmp(p, si[s].desc, strlen(p)) == 0)
			break;
		if (s >= nisig)
		    (void)fprintf(stderr,
			   "%s: (warning) no signal named '%s' in record %s\n",
				  pname, p, record);
	    }
	    if (s < nisig)
		sig[j++] = s;
	}
	if (j < 1) {
	    (void)fprintf(stderr, "%s: no usable signals in record %s\n",
			  pname, record);
	    cleanup(4);
	}
	nisig = j;
    }
    else {

#ifndef MULTILEAD
	nisig = 1;			/* analyze first signal only */
#endif

	sig = (WFDB_Signal *)gcalloc((size_t)nisig, sizeof(int));
	for (i = 0; i < nisig; i++)
	    sig[i] = i;
	nsig = nisig;
    }

    vo = (WFDB_Sample *)gcalloc((size_t)nsig, sizeof(WFDB_Sample));

    if (orec) {			/* save output signal(s) in a new record */
        so = (WFDB_Siginfo *)gcalloc((size_t)nsig, sizeof(WFDB_Siginfo));
	osfname = (char *)gcalloc((size_t)(strlen(orec)+5), sizeof(char));
        sprintf(osfname, "%s.dat", orec);
        for (i = 0; i < nsig; i++) {
	    so[i] = si[sig[i]];
	    so[i].fname = osfname;
	    so[i].group = 0;
	    so[i].fmt = 16;
	    so[i].spf = 1;
	}
        if (osigfopen(so, nsig) < 0)
	    cleanup(2);
    }

    /* Record the argument list in a NOTE annotation. */
    for (i = j = 0, p = auxbuf+1; i < argc; i++) {
	int len;
	j += len = strlen(argv[i]);
	if (j < 255) {
	    strncpy(p, argv[i], len);
	    p += len;
	    *p++ = ' ';
	    j++;
	}
	else {
	    j -= len;
	    break;
	}
    }	
    *(--p) = '\0';
    auxbuf[0] = j;
    annot.subtyp = annot.chan = annot.num = 0;
    annot.aux = auxbuf;
    annot.anntyp = NOTE;
    annot.time = (WFDB_Time)0;
    putann(0, &annot);
    annot.anntyp = NORMAL;
    annot.aux = NULL;	

    gqrs_init(from, to);	/* initialize variables for gqrs() */
    gqrs(from, to);		/* run the detector and collect output */

    printf(" %s\n", timstr(strtim("i")));
    cleanup(0);	/*NOTREACHED*/
}

int dt, dt2, dt3, dt4, smdt;	/* time intervals set by gqrs_init() depending
				   on the sampling frequency and used by
				   filter functions sm() and qf();  units
				   are sample intervals */
WFDB_Time smt = 0, smt0;	/* current and starting time for sm() */
long *v1, **qfv, **smv;		/* filter buffers allocated by gqrs_init() */

/* The q() and s() macros can be used for fast lookup of filter outputs.  Note
   that the first arguments of these macros, and of functions sm() and qf(),
   are the *mapped* signal numbers (i.e., 0, 1, ...), which do not necessarily
   match the input signal numbers sig[0], sig[1], ... . */

#define q(C,T) (qfv[(C)][(T)&(BUFLN-1)])
#define s(C,T) (smv[(C)][(T)&(BUFLN-1)])

/* sm() implements a trapezoidal low pass (smoothing) filter (with a gain of
   4*smdt) applied to the ECG signals before the QRS matched filter qf().
   Before attempting to 'rewind' by more than BUFLN-smdt samples, reset smt
   and smt0.
 */

WFDB_Sample sm(WFDB_Signal c, WFDB_Time t)
{
    while (t > smt) {
	int i;
	WFDB_Signal c;

	if (++smt > smt0)	/* fast update by summing first differences */
	    for (i = 0; i < nsig; i++) {
		c = sig[i];
		s(i, smt) = s(i, smt-1) +
		    sample(c, smt+smdt) + sample(c, smt+smdt-1) -
		    sample(c, smt-smdt) - sample(c, smt-smdt-1);
	    }
	else			/* get initial values by full convolution */
	    for (i = 0; i < nsig; i++) {
		int j, v;

		c = sig[i];
		for (j = 1, v = sample(c, smt); j < smdt; j++)
		    v += sample(c, smt+j) + sample(c, smt-j);
		s(i, smt) = (v << 1) + sample(c, smt+j) + sample(c, smt-j)
		    - si[c].adczero * (smdt << 2); /* FIXME: is this needed? */
	    }
    }
    return (s(c, t));
}

void qf()	/* evaluate the QRS detector filter for the next sample */
{
    int c;

    for (c = 0; c < nsig; c++) {
	long dv, dv1, dv2, v0;

	dv2 = sm(c, t+dt4);/* do this first, to ensure that all of the other
			      smoothed values needed below are in the buffer */
	dv2 -= s(c, t-dt4);
	dv1 = s(c, t+dt)  - s(c, t-dt);
       	dv  = (dv1 << 1);
	dv -= s(c, t+dt2) - s(c, t-dt2);
	dv <<= 1;
	dv += dv1;
	dv -= s(c, t+dt3) - s(c, t-dt3);
	dv <<= 1;
	dv += dv2;
	v1[c] += dv;
	v0 = v1[c] / v1norm;  /* scaling is needed to avoid overflow */
	q(c, t) = vo[c] = v0 * v0;
    }
}

struct peak **peaks, **cpeak;

void addpeak(WFDB_Signal c, WFDB_Time t, int peak_amplitude)
{
    struct peak *p = cpeak[c]->next;

    p->time = t;
    p->amp = peak_amplitude;
    p->type = 0;
    cpeak[c] = p;
    (p->next)->amp = 0;
}

/* peaktype() returns 1 if p is the most prominent peak in its neighborhood, 2
   otherwise.  The neighborhood consists of all other peaks within rrmin.
   Normally, "most prominent" is equivalent to "largest in amplitude", but this
   is not always true.  For example, consider three consecutive peaks a, b, c
   such that a and b share a neighborhood, b and c share a neighborhood, but a
   and c do not; and suppose that amp(a) > amp(b) > amp(c).  In this case, if
   there are no other peaks, a is the most prominent peak in the (a, b)
   neighborhood.  Since b is thus identified as a non-prominent peak, c becomes
   the most prominent peak in the (b, c) neighborhood.  This is necessary to
   permit detection of low-amplitude beats that closely precede or follow beats
   with large secondary peaks (as, for example, in R-on-T PVCs).
*/

int peaktype(struct peak *p)
{
    if (p->type)
	return (p->type);
    else {
	int a = p->amp;
	struct peak *pp;
	WFDB_Time t0 = p->time - rrmin, t1 = p->time + rrmin;

	if (t0 < 0) t0 = 0;
	for (pp = p->prev; t0 < pp->time && pp->time < (pp->next)->time;
	     pp = pp->prev) {
	    if (pp->amp == 0) break;
	    if (a < pp->amp && peaktype(pp) == 1)
		return (p->type = 2);
	}
	for (pp = p->next; pp->time < t1 && pp->time > (pp->prev)->time;
	     pp = pp->next) {
	    if (pp->amp == 0) break;
	    if (a < pp->amp && peaktype(pp) == 1)
		return (p->type = 2);
	}
	return (p->type = 1);
    }
}

/* gqrs_init() is intended to be called once only, to initialize variables for
   the QRS detection function gqrs(). */

void gqrs_init(WFDB_Time from, WFDB_Time to)
{
    int c;
    static double HR, RR, RRdelta, RRmin, RRmax, QS, QT, RTmax, QRSa, QRSamin;
    
    /* Allocate workspace.  The allocator, gcalloc, is defined below;  it
       includes an error handler and on-exit deallocation so that this
       code can be kept simpler. */
    v1 = (long *)gcalloc((size_t)nsig, sizeof(long));
    pthr = (int *)gcalloc((size_t)nsig, sizeof(int));
    qthr = (int *)gcalloc((size_t)nsig, sizeof(int));
    pthmin = (int *)gcalloc((size_t)nsig, sizeof(int));
    qthmin = (int *)gcalloc((size_t)nsig, sizeof(int));
    tamean = (int *)gcalloc((size_t)nsig, sizeof(int));
    qfv = (long **)gcalloc((size_t)nsig, sizeof(long *));
    smv = (long **)gcalloc((size_t)nsig, sizeof(long *));
    peaks = (struct peak **)gcalloc((size_t)nsig, sizeof(struct peak *));
    cpeak = (struct peak **)gcalloc((size_t)nsig, sizeof(struct peak *));
    for (c = 0; c < nsig; c++) {
	int i;

	qfv[c] = (long *)gcalloc((size_t)BUFLN, sizeof(long));
	smv[c] = (long *)gcalloc((size_t)BUFLN, sizeof(long));
	peaks[c] = (struct peak *)gcalloc((size_t)NPEAKS, sizeof(struct peak));
	/* Gather peak structures into circular buffers */
	for (i = 0; i < NPEAKS; i++) {
	    peaks[c][i].next = &peaks[c][i+1];
	    peaks[c][i].prev = &peaks[c][i-1];
	}
	peaks[c][0].prev = &peaks[c][NPEAKS-1];
	cpeak[c] = peaks[c][NPEAKS-1].next = &peaks[c][0];
    }

    /* Read a priori physiologic parameters from the configuration file if
       available. They can be adjusted in the configuration file for pediatric,
       fetal, or animal ECGs. */
    if (config) {
	char buf[256], *p;

	/* Read the configuration file a line at a time. */
	while (fgets(buf, sizeof(buf), config)) {
	    /* Skip comments (empty lines and lines beginning with `#'). */
	    if (buf[0] == '#' || buf[0] == '\n') continue;
	    /* Set parameters.  Each `getconf' below is executed once for
	       each non-comment line in the configuration file. */
	    getconf(HR, "%lf");
	    getconf(RR, "%lf");
	    getconf(RRdelta, "%lf");
	    getconf(RRmin, "%lf");
	    getconf(RRmax, "%lf");
	    getconf(QS, "%lf");
	    getconf(QT, "%lf");
	    getconf(RTmax, "%lf");
	    getconf(QRSa, "%lf");
	    getconf(QRSamin, "%lf");
	}
	fclose(config);
    }

    /* If any a priori parameters were not specified in the configuration file,
       initialize them here (using values chosen for adult human ECGs). */
    if (HR != 0.0) RR = 60.0/HR;
    if (RR == 0.0) RR = 0.8;
    if (RRdelta == 0.0) RRdelta = RR/4;
    if (RRmin == 0.0) RRmin = RR/4;
    if (RRmax == 0.0) RRmax = 3*RR;
    if (QS == 0.0) QS = 0.07;
    if (QT == 0.0) QT = 5*QS;
    if (RTmax == 0.0) RTmax = 5*QS;
    if (QRSa == 0.0) QRSa = 750;
    if (QRSamin == 0.0) QRSamin = QRSa/5;

    /* Initialize gqrs's adaptive parameters based on the a priori parameters.
       During its learning period, gqrs will adjust them based on the observed
       input; after the learning period, gqrs continues to adjust these
       parameters, but with slower rates of change than during the learning
       period. */
    rrmean = RR * sps;
    rrdev = RRdelta * sps;
    rrmin = RRmin * sps;
    rrmax = RRmax * sps;
    if ((dt = QS * sps / 4) < 1) {
	dt = 1;
	fprintf(stderr, "%s (warning): sampling rate may be too low\n", pname);
    }
    rtmean = 0.75 * QT * sps;	/* expected RTpeak interval, about 75% of QT */
    rtmax = RTmax * sps;	/* maximum RTpeak interval */
    for (c = 0; c < nsig; c++) {
	int dv = muvadu(sig[c], (int)(QRSamin));

	pthr[c] = (thresh * dv * dv) / 6;
	qthr[c] = pthr[c] << 1;
	pthmin[c] = pthr[c] >> 2;
	qthmin[c] = (pthmin[c] << 2)/3;
	tamean[c] = qthr[c];	/* initial value for mean T-wave amplitude */
    }

    /* Filter constants and thresholds. */
    dt2 = 2*dt;
    dt3 = 3*dt;
    dt4 = 4*dt;
    smdt = dt;
    smt = from;
    smt0 = from + smdt;
    v1norm = smdt * dt * 64;

    if ((rrinc = rrmean/40) < 1) rrinc = 1;    
    t -= dt4;
    (void)sample(0,t);

    state = RUNNING;
}

/* find_missing() is invoked by gqrs() whenever it is suspected that a
   low-amplitude beat may have been missed between two consecutive detected
   beats at r and p.  The primary peak closest to the expected time of
   the missing beat, if any, is returned as the suggested missing beat. */

struct peak *find_missing(struct peak *r, struct peak *p)
{
    int rrerr, rrtmp, minrrerr;
    struct peak *q, *s = NULL;

    if (r == NULL || p == NULL) return (NULL);
    minrrerr = p->time - r->time;
    for (q = r->next; q->time < p->time; q = q->next) {
	if (peaktype(q) == 1) {
	    rrtmp = q->time - r->time;
	    rrerr = rrtmp - rrmean;
	    if (rrerr < 0) rrerr = -rrerr;
	    if (rrerr < minrrerr) {
		minrrerr = rrerr;
		s = q;
	    }
	}
    }
    return (s);
}

/* gqrs() is the main QRS detection function.  It attempts to find all
   beats between the time limits specified by its arguments, and to label
   them using annotations of type NORMAL (gqrs() does not attempt to
   differentiate normal and ectopic beats). */

void gqrs(WFDB_Time from, WFDB_Time to)
{
    int c, i, qamp, q0, q1 = 0, q2 = 0, rr, rrd, rt, rtd, rtdmin;
    struct peak *p, *q = NULL, *r = NULL, *tw;
    WFDB_Time last_peak = from, last_qrs = from;

    while (to == (WFDB_Time)0 || t <= to) {
	static int countdown = -1;

	if (countdown < 0 && sample_valid())
	    qf();
	else if (countdown < 0) {
	    countdown = strtim("1");
	    state = CLEANUP;
	}
	else if (countdown-- <= 0)
	    break;

	for (c = 0; c < nsig; c++) {
	q0 = q(c, t); q1 = q(c, t-1); q2 = q(c, t-2);
	if (q1 > pthr[c] && q2 < q1 && q1 >= q0 && t > dt4) {
	    addpeak(c, t-1, q1);
	    last_peak = t-1;
	    for (p = cpeak[c]->next; p->time < t - rrmin; p = p->next) {
		if (p->time >= annot.time + rrmin && peaktype(p) == 1) {
		    if (p->amp > qthr[c]) {
			rr = p->time - annot.time;
			if (rr > rrmean + 2 * rrdev &&
			    rr > 2 * (rrmean - rrdev) &&
			    (q = find_missing(r, p))) {
			    p = q;
			    rr = p->time - annot.time;
			    annot.subtyp = 1;
			}
			if (c == 0) {
			    if ((rrd = rr - rrmean) < 0) rrd = -rrd;
			    rrdev += (rrd - rrdev) >> 3;
			    if (rrd > rrinc) rrd = rrinc;
			    if (rr > rrmean) rrmean += rrd;
			    else rrmean -= rrd;
			    if (p->amp > qthr[c] * 4) qthr[c]++;
			    else if (p->amp < qthr[c]) qthr[c]--;
			    if (qthr[c] > pthr[c] * 20) qthr[c] = pthr[c] * 20;
			    last_qrs = p->time;
			}
			if (state == RUNNING) {
			    int qsize;

			    annot.anntyp = NORMAL;
			    annot.chan = c;
			    annot.time = p->time - dt2;
			    qsize = p->amp * 10.0 / qthr[c];
			    if (qsize > 127) qsize = 127;
			    annot.num = qsize;
			    putann(0, &annot);
			    annot.time += dt2;
			}
			/* look for this beat's T-wave */
			tw = NULL; rtdmin = rtmean;
			for (q = p->next; q->time > annot.time; q = q->next) {
			    rt = q->time - annot.time;
			    if ((rtd = rt - rtmean) < 0) rtd = -rtd;
			    if (rtd < rtdmin) {
				rtdmin = rtd;
				tw = q;
			    }
			    if (q->time > annot.time + rtmax)
				break;
			}
			if (tw) {
			    rt = tw->time - annot.time;
			    if ((rtmean += (rt - rtmean) >> 4) > rtmax)
				rtmean = rtmax;
			    tw->type = 2;	/* mark T-wave as secondary */
			}
			r = p; q = NULL; qamp = 0; annot.subtyp = 0;
		    }
		    else if (t - last_qrs > rrmax && qthr[c] > qthmin[c])
			qthr[c] -= (qthr[c] >> 4);
		}
	    }
	}
	else if (t - last_peak > rrmax)
	    for (c = 0; c < nsig; c++)
		if (pthr[c] > pthmin[c])
		    pthr[c] -= (pthr[c] >> 4);

	if (state == RUNNING && orec && t >= 0) putvec(vo);

	if (t >= next_minute) {
	    next_minute += spm;
	    (void)fprintf(stderr, ".");
	    (void)fflush(stderr);
	    if (++minutes >= 60) {
		(void)fprintf(stderr, " %s\n", timstr(-t));
		minutes = 0;
	    }
	}
      }
      t++;
    }
    
    /* Mark the last beat or two. */
    for (c = 0; c < nsig; c++)
	for (p = cpeak[c]->next; p->time < (p->next)->time; p = p->next) {
	    if (to > 0 && p->time > to - sps)
		break;
	    if (p->time >= annot.time + rrmin && peaktype(p) == 1) {
		annot.anntyp = (c == 0) ? NORMAL : ARFCT;
		annot.chan = c;
		annot.time = p->time;
		putann(0, &annot);
	    }
	}
}

/* prog_name() extracts this program's name from argv[0], for use in error and
   warning messages. */

char *prog_name(char *s)
{
    char *p = s + strlen(s);

#ifdef MSDOS
    while (p >= s && *p != '\\' && *p != ':') {
	if (*p == '.')
	    *p = '\0';		/* strip off extension */
	if ('A' <= *p && *p <= 'Z')
	    *p += 'a' - 'A';	/* convert to lower case */
	p--;
    }
#else
    while (p >= s && *p != '/')
	p--;
#endif
    return (p+1);
}

/* help() prints a (very) concise summary of how to use this program.
   A more detailed summary is in the man page (gqrs.1). */

static char *help_strings[] = {
 "usage: %s -r RECORD [OPTIONS ...]\n",
 "where RECORD is the name of the record to be analyzed, and OPTIONS may",
 "include any of:",
 " -c FILE     initialize parameters from the specified configuration FILE",
 " -f TIME     begin at specified time",
 " -h          print this usage summary",
 " -H          read multifrequency signals in high resolution mode",
 " -m THRESH   set detector threshold to THRESH (default: 1.00)",
 " -o RECORD   save filtered signals in a new RECORD", 
 " -s SIGNAL [ SIGNAL ] ...   analyze specified signals (default: 0)",
 "                (Note: SIGNALs may be specified by number or name.)",
#ifndef MULTILEAD
 "                (This version analyzes only the first specified signal!)",
#endif
 " -t TIME     stop at specified time",
 "If too many beats are missed, decrease THRESH;  if there are too many extra",
 "detections, increase THRESH.",
NULL
};

void help()
{
    int i;

    (void)fprintf(stderr, help_strings[0], pname);
    for (i = 1; help_strings[i] != NULL; i++)
	(void)fprintf(stderr, "%s\n", help_strings[i]);
}

/* gcalloc() is a wrapper for calloc() that handles errors and maintains
   a list of allocated buffers for automatic on-exit deallocation via
   gfreeall(). */

size_t gcn = 0, gcnmax = 0;
void **gclist = NULL;

void *gcalloc(size_t nmemb, size_t size)
{
    void *p = calloc(nmemb, size), **q = NULL;

    if ((p == NULL) ||
	((gcn >= gcnmax) &&
	 (q = realloc(gclist, (gcnmax += 32)*(sizeof(void *)))) == NULL)) {
	fprintf(stderr, "%s: insufficient memory\n", pname);
	cleanup(3);
    }
    if (q) gclist = q;
    return (gclist[gcn++] = p);
}

void gfreeall()			/* free memory allocated using gcalloc() */
{
   while (gcn-- > 0)
	if (gclist[gcn]) free(gclist[gcn]);
    free(gclist);
}

void cleanup(int status)	/* close files and free allocated memory */
{
    if (orec) newheader(orec);
    if (record) wfdbquit();
    gfreeall();
    exit(status);
}
